#if !defined(FENVCW_H)
#define FENVCW_H

/***********************************************************************
This header file is a replacement for, and interface to, <fenv.h>.

Most pre-C99 systems, apart from GNU/Linux, lack <fenv.h>.  On such
systems, this header file defines suitable values of the FE_xxx flags
for access to precision control:

	FE_FLTPREC      FE_DBLPREC      FE_LDBLPREC

rounding modes:

	FE_DOWNWARD     FE_TONEAREST    FE_TOWARDZERO   FE_UPWARD

sticky exception flags:

	FE_ALL_EXCEPT   FE_INEXACT      FE_INVALID      FE_SUBNORMAL
	FE_UNDERFLOW    FE_DIVBYZERO    FE_OVERFLOW

(of which FE_SUBNORMAL is an extension supported by some, but not all,
floating-point architectures), via the 14 functions (their prototypes
appear below):

	feclearexcept	  fegetround	    fesetenv	      fesetround
	fegetenv	  feholdexcept	    fesetexceptflag   fetestexcept
	fegetexceptflag	  feraiseexcept	    fesetprec	      feupdateenv
	fegetprec

with these additional opaque data types:

	fenv_t		fexcept_t

and this opaque value:

	FE_DFL_ENV

This makes a uniform C99-style interface available to code running on
older systems, or compiled with older compilers.

On systems that have <fenv.h>, defined at compile time by the symbol
HAVE_FENV_H (automatically set by C99-mode compilation), this header
file simply includes the system one, and turns on the FENV_ACCESS
pragma, which warns the compiler and optimizer that the code may get
and set hardware flags, and thus, cannot allow certain kinds of code
reordering.

Thus, use
	#include "fenvcw.h"
instead of
	#include <fenv.h>
and link with the mathcw library (on Unix systems, -lmcw) to get the
interface library support.  A different header file name is essential,
since this file may include <fenv.h>.

Compiling with a C99 compiler (e.g., c99, gcc -stdc=c99, icc
-stdc=c99, pathcc -stdc=c99, ...) will use the native header file and
library.

Care must be taken not to mix code compiled with and without <fenv.h>,
since the two data types are not guaranteed to match between
implementations.  Their structure is opaque to the user, but
nevertheless, they must be used consistently.

***********************************************************************/

#if defined(__STDC_VERSION__) && ((__STDC_VERSION__ + 0) >= 199901L)
#if defined(__FreeBSD__)
    /* No <fenv.h> on this system */
#elif defined(__NetBSD__)
    /* No <fenv.h> on this system */
#elif defined(__OpenBSD__)
    /* No <fenv.h> on this system */
#elif defined(__sgi__) && defined(__UPC__)
    /* No <fenv.h> on this system */
#else
#define HAVE_FENV_H
#endif
#endif /* defined(__STDC_VERSION__) && ((__STDC_VERSION__ + 0) >= 199901L) */

#if defined(__linux) && !defined(HAVE_FENV_H)
#define HAVE_FENV_H
#endif

#if defined(__SunOS_5_10) && !defined(HAVE_FENV_H)
#define HAVE_FENV_H
#endif

#if defined(__SunOS_5_11) && !defined(HAVE_FENV_H)
#define HAVE_FENV_H
#endif

#if defined(__FreeBSD__)	/* <fenv.h> available from version 6 onward */
#if (__FreeBSD__ + 0) >= 6
#define HAVE_FENV_H
#endif
#endif /* defined(__FreeBSD__) */

/**********************************************************************/
#if defined(HAVE_FENV_H)
			/* C99 style, now supported by many systems,
		   	   although sometimes extra libraries are needed */
#include <fenv.h>
#pragma STDC FENV_ACCESS ON	/* required by 1999 ISO C Standard */

#else /* not defined(HAVE_FENV_H) */

#include <stddef.h>		/* for NULL macro */

#if defined(__cplusplus)
extern "C" {
#endif

typedef struct
{				/* pad in case of mathcw compile and -lm link */
    int fe_sticky_flags;
    int unused_padding[63];	/* longest known native fenv_t is 100 bytes */
} fenv_t;

typedef struct
{
    int fe_sticky_flags;
    int unused_padding[3];	/* longest known native fexcept_t is 8 bytes */
} fexcept_t;

extern const fenv_t __mcw_fe_dfl_env;

#if defined(__COMPILER_KCC__)
#define HAVE_SHORT_NAMES	1
#endif

#if defined(HAVE_SHORT_NAMES)
#include <shrtcw.h>
#endif /* defined(HAVE_SHORT_NAMES) */

#define	FE_DFL_ENV	(&__mcw_fe_dfl_env)

extern int	(feclearexcept)(int);
extern int	(fegetenv)(fenv_t *);
extern int	(fegetexcept)(void);
extern int	(fegetexceptflag)(fexcept_t *, int);
extern int	(fegetround)(void);
extern int	(feholdexcept)(fenv_t *);
extern int	(feraiseexcept)(int);
extern int	(fesetenv)(const fenv_t *);
extern int	(fesetexceptflag)(const fexcept_t *, int);
extern int	(fesetround)(int);
extern int	(fetestexcept)(int);
extern int	(feupdateenv)(const fenv_t *);

#if defined(__cplusplus)
};
#endif

#if defined(HAVE_OS_OSF) || defined(__osf__) || \
	(defined(HAVE_FLOAT_H) && defined(HAVE_READ_RND) && defined(HAVE_WRITE_RND))
				/* OSF/1 style */
#include <float.h>

#if !defined(FP_RND_RZ)		/* happens with lcc strict Standard C <float.h> */
#define FP_RND_RZ       0
#define FP_RND_RN       1
#define FP_RND_RP       2
#define FP_RND_RM       3
#endif /* !defined(FP_RND_RZ) */

#define FE_DOWNWARD		FP_RND_RM
#define FE_TONEAREST		FP_RND_RN
#define FE_TOWARDZERO		FP_RND_RZ
#define FE_UPWARD		FP_RND_RP

#if !defined(FE_ALL_EXCEPT)
#define FE_DENORMAL		(1UL << 22)
#define FE_INEXACT		(1UL << 21)
#define FE_UNDERFLOW		(1UL << 20)
#define FE_OVERFLOW		(1UL << 19)
#define FE_DIVBYZERO		(1UL << 18)
#define FE_INVALID		(1UL << 17)
#define FE_ALL_EXCEPT		(0x3f << 17)
#endif /* !defined(FE_ALL_EXCEPT) */

/**********************************************************************/
#elif defined(HAVE_OS_AIX) || defined(_AIX) || \
	(defined(HAVE_FLOAT_H) && defined(HAVE_WRITE_RND) && defined(HAVE_FP_READ_RND))
				/* IBM AIX 4.x style */
#include <float.h>
#define FE_DOWNWARD		FP_RND_RM
#define FE_TONEAREST		FP_RND_RN
#define FE_TOWARDZERO		FP_RND_RZ
#define FE_UPWARD		FP_RND_RP

#define FE_DIVBYZERO		0x01
#define FE_INEXACT		0x02
#define FE_INVALID		0x04
#define FE_OVERFLOW		0x08
#define FE_UNDERFLOW		0x10

#define FE_ALL_EXCEPT		(FE_DIVBYZERO | FE_INEXACT | FE_INVALID | FE_OVERFLOW | FE_UNDERFLOW)

/**********************************************************************/
#elif !defined(__vax__) && \
      ((defined(HAVE_OS_IRIX) || \
	defined(HAVE_OS_FREEBSD) || \
	defined(HAVE_OS_NETBSD) || \
	defined(HAVE_OS_OPENBSD)) || \
       (defined(__sgi) || \
	defined(__FreeBSD__) || \
	defined(__NetBSD__) || \
	defined(__OpenBSD__)) || \
       (defined(HAVE_IEEEFP_H) && \
	defined(HAVE_FPGETROUND) && \
	defined(HAVE_FPSETROUND)))
				/* IRIX, FreeBSD, NetBSD, OpenBSD style */

#if defined(HAVE_OS_NETBSD) && !defined(FE_INEXACT)

#define	FE_INVALID	0x01	/* invalid operation exception */
#define	FE_DENORMAL	0x02	/* denormalization exception */
#define	FE_DIVBYZERO	0x04	/* divide-by-zero exception */
#define	FE_OVERFLOW	0x08	/* underflow exception */
#define	FE_UNDERFLOW	0x10	/* overflow exception */
#define	FE_INEXACT	0x20	/* imprecise (loss of precision) */

#define	FE_ALL_EXCEPT	0x3f

#define	FE_TONEAREST	0	/* round to nearest representable number */
#define	FE_DOWNWARD	1	/* round toward negative infinity */
#define	FE_UPWARD	2	/* round toward positive infinity */
#define	FE_TOWARDZERO	3	/* round to zero (truncate) */

#define FP_X_INV	FE_INVALID	/* invalid operation exception */
#define FP_X_DNML	FE_DENORMAL	/* denormalization exception */
#define FP_X_DZ		FE_DIVBYZERO	/* divide-by-zero exception */
#define FP_X_OFL	FE_OVERFLOW	/* overflow exception */
#define FP_X_UFL	FE_UNDERFLOW	/* underflow exception */
#define FP_X_IMP	FE_INEXACT	/* imprecise (loss of precision) */

#define HAVE_BROKEN_IEEEFP_H

typedef int fp_except;

typedef enum
{
    FP_RN=0,
    FP_RM=1,
    FP_RP=2,
    FP_RZ=3
} fp_rnd;

#else

#include <ieeefp.h>

#endif /* defined(HAVE_OS_NETBSD) && !defined(FE_INVALID) */

#if defined(HAVE_FP_EXCEPT_T) && !defined(HAVE_FP_EXCEPT)
#define fp_except fp_except_t 	/* map more common type to FreeBSD type */
#endif

#if defined(HAVE_FP_RND_T) && !defined(HAVE_FP_RND)
#define fp_rnd fp_rnd_t		/* map more common type to FreeBSD type */
#endif

#if !defined(FE_ALL_EXCEPT)

#define FE_DOWNWARD		FP_RM
#define FE_TONEAREST		FP_RN
#define FE_TOWARDZERO		FP_RZ
#define FE_UPWARD		FP_RP

#define FE_DIVBYZERO		FP_X_DZ
#define FE_INEXACT		FP_X_IMP
#define FE_INVALID		FP_X_INV
#define FE_OVERFLOW		FP_X_OFL
#define FE_UNDERFLOW		FP_X_UFL

#if defined(FP_X_DNML)
#define FE_SUBNORMAL		FP_X_DNML /* extension to C99 supported on IA-32 */
#else
#define FE_SUBNORMAL		(FE_ALL_EXCEPT + 1)
#endif

#define FE_ALL_EXCEPT		(FE_DIVBYZERO | FE_INEXACT | FE_INVALID | FE_OVERFLOW | FE_UNDERFLOW)

#endif

/**********************************************************************/
#elif defined(HAVE_OS_HPUX) || defined(__hpux) || \
	(defined(HAVE_MATH_H) && defined(HAVE_FPGETROUND) && defined(HAVE_FPSETROUND))
				/* HP-UX 10 style */
#include <math.h>

#define FE_DOWNWARD		FP_RM
#define FE_TONEAREST		FP_RN
#define FE_TOWARDZERO		FP_RZ
#define FE_UPWARD		FP_RP

#define FE_DIVBYZERO		0x01
#define FE_INEXACT		0x02
#define FE_INVALID		0x04
#define FE_OVERFLOW		0x08
#define FE_UNDERFLOW		0x10

#define FE_ALL_EXCEPT		(FE_DIVBYZERO | FE_INEXACT | FE_INVALID | FE_OVERFLOW | FE_UNDERFLOW)

/**********************************************************************/
#else
				/* No floating-point control available.
				   Functions are silent, or have error returns. */
#if !defined(FE_DOWNWARD)
#define FE_DOWNWARD		3
#define FE_TONEAREST		0
#define FE_TOWARDZERO		1
#define FE_UPWARD		2
#endif

#if !defined(FE_DIVBYZERO)
#define FE_DIVBYZERO		0x01
#define FE_INEXACT		0x02
#define FE_INVALID		0x04
#define FE_OVERFLOW		0x08
#define FE_UNDERFLOW		0x10

#define FE_ALL_EXCEPT		(FE_DIVBYZERO | FE_INEXACT | FE_INVALID | FE_OVERFLOW | FE_UNDERFLOW)
#endif

#endif /* defined(HAVE_OS_OSF) ... */

#endif /* defined(HAVE_FENV_H) */

#if defined(__cplusplus)
extern "C" {
#endif

extern int	(fegetprec)(void);
extern int	(fesetprec)(int);

#if defined(__cplusplus)
}
#endif

#if !defined(FE_FLTPREC)	/* already defined in <fenv.h> only on Sun Solaris x86 */
#define FE_FLTPREC      0
#define FE_DBLPREC      2
#define FE_LDBLPREC     3
#endif

/* Additional architecture-specific exception flags (zero if not available) */

#undef FE_SUBNORMAL

#if defined(FE_DENORMAL)	/* GNU/Linux on Alpha */
#define FE_SUBNORMAL		FE_DENORMAL
#elif defined(__FE_DENORM)	/* GNU/Linux on AMD64, IA-32 */
#define FE_SUBNORMAL		__FE_DENORM
#elif defined(FE_UNNORMAL)	/* GNU/Linux on IA-64 */
#define FE_SUBNORMAL		FE_UNNORMAL
#elif defined(__alpha__)	/* GNU/Linux and OSF/1 on Alpha */
#define FE_SUBNORMAL		(1UL << 22)
#elif defined(__i386__)		/* GNU/Linux on IA-32 */
#define FE_SUBNORMAL		0x02
#elif defined(__x86_64__)	/* GNU/Linux on AMD64 */
#define FE_SUBNORMAL		0x02
#else  /* apparently not supported on this platform */
    /* C99 Standard requires that unsupported flags not be defined */
#endif

#endif /* !defined(FENVCW_H) */
